#
# Copyright (c) 1998 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# ------------
#
# Filename: application-mui.tcl
#   -- Author: Cynthia Romer <cromer@cs.berkeley.edu>
#
#  @(#) $Header: /usr/src/mash/repository/mash/mash-1/tcl/mui/application-mui.tcl,v 1.30 1998/07/01 22:38:20 schuett Exp $
#

import RTPApplication Configuration AddressBlock VideoHandler \
		AudioAgent MuiUI \
		FontInitializer WidgetResourceInitializer \
		CoordinationBus \
		MTrace Tcl_Player Tcl_Recorder NetworkManager \
		MBDebugDlg MBUI MBPSInterp MB_Manager MBPageMgr MBNet \
		Program UserApplication/MUI SDPParser MPlug/ScrolledUI 

#
# An Application for a collaborative video, audio, and shared whiteboard
# tool.  (Although this application supports running any of its media
# with MeGa, nsdr's UserApplication/MUI::match method has yet to be
# modified to supply the necessary command-line options.)
#
Class MuiApplication -superclass RTPApplication

#
# Invoked with "new MuiApplication".  Parse the options supplied by
# <i>argv</i>.  Creates a frame using the provided <i>widgetPath</i> and
# pack the mui application interface in it.
#
MuiApplication public init { widgetPath argv } {
	$self next mui

	# Initialization of variables and resource db options
	set o [$self options]
	$self init_args $o

	$self init_fonts $o

	# load user prefs before process argv
	$o load_preferences "rtp mui"
	set argv [$o parse_args $argv]
	append argv [$self parse_sdp_msg]
	# Generate an error for extra args.
	$self check_argv $argv

	# Retrieving and checking for at least one media spec.
	$self instvar vspec_ aspec_ mbspec_	
	set vspec_ [$self get_option videoSessionSpec] 
	set aspec_ [$self get_option audioSessionSpec] 
	set mbspec_ [$self get_option mbSessionSpec] 
	$self instvar have_video_ have_audio_ have_mb_
	$self check_hostspecs
	
	$self check_rtp_sdes

	# init resources & load preferences
	new WidgetResourceInitializer
	if { $have_video_ } {
		$self init_vic_resources $o
	}
	if { $have_audio_ } {
		$self init_vat_resources $o
	}
	if { $have_mb_ } {
		$self init_mb_resources $o
	}
	$self init_mui_resources $o

	# Backward compat. -B flag is kb/s.
	set t [$self get_option maxbw]
	if { $t > 0 } {
		$self add_option maxbw [expr 1000*$t]
	}

	# Create a Coordination Bus
	$self init_confbus

	if { $have_video_ } {
		# Create a VideoAgent, VideoPipeline, and possibly set them up to use SCUBA
		$self instvar videoAgent_ vpipe_ videoHandler_ scuba_sess_	
		set videoHandler_ [new VideoHandler $vspec_]
		set videoAgent_ [$videoHandler_ agent]
		set vpipe_ [$videoHandler_ vpipe]
		if { [$self get_option useScuba] != "" } {
			set scuba_sess_ [$videoHandler_ set scuba_sess_]
		}
		#XXX need to do this to size the bandwidth slider in the GUI
		$self add_option maxbw [[new AddressBlock $vspec_] set maxbw_(0)]
	}

	if { $have_audio_ } {
		# Create AudioAgent and put it on the conference bus.
		$self instvar audioAgent_ local_chan_ glob_chan_
		set audioAgent_ [new AudioAgent $self $aspec_]
		$audioAgent_ attach_local_channel $local_chan_
		$audioAgent_ attach_global_channel $glob_chan_
	}

	
	if { $have_mb_ } {
		MTrace init {trcMB}
		# must do this before create ui, because mbManager is required to make ui
		$self init_mb_variables
	}

	# Create the user interface for this application.
	$self instvar ui_ mbui_
	set ui_ [$self init_ui $widgetPath $vspec_]
	
	if { $have_mb_ } {
		$self instvar mgr_
		# must create ui before do this, because need mbui arg
		set mbnet [new MBNet $mbui_ $mgr_]
		$self instvar session_ mbAgent_
		set session_ [$mbnet session]
		set mbAgent_ [$mbnet mbAgent]
		$mbnet adjust_spec
		
		set logf [$o get_option playScript]
		if {"none" != $logf} {
			global mb
			set isRT [$config get_option rtPlay]
			set player [new Tcl_Player $logf [$self set sender_] $isRT]
			# wait for 1 second for everything to settle down
			after 15000 $player start
		}
	}
	
	$self user_hook
}

#
# Create the user interface for this application.
# Creates a frame using the provided <i>widgetPath</i> and pack the mui ui in it.
#
MuiApplication public init_ui { widgetPath {vSpec {}} } {
	$self instvar videoAgent_ audioAgent_ vpipe_ scuba_sess_
        $self instvar local_chan_ glob_chan_ mbui_ mgr_ sender_

	set w $widgetPath
	frame $w

	$self instvar have_video_ have_audio_ have_mb_
	set video_args {}
	set audio_args {}
	set mb_args {}
	if $have_video_ {
		lappend video_args $videoAgent_
		lappend video_args $vpipe_
	}
	if $have_audio_ {
		lappend audio_args $audioAgent_
	}
	if $have_mb_ {
		lappend mb_args $mgr_
		lappend mb_args $sender_
	}
	
	set ui [new MuiUI $w $local_chan_ $glob_chan_ $video_args $audio_args $mb_args "$self exit" $vSpec]
	
	if $have_mb_ {
		set mbui_ [$ui mbui]
	}
	
	# move this to MuiUI init, just to prove that it can be done	
	if {0} {
		import VicUI VatUI MBUI
		$self instvar local_chan_ glob_chan_

		if {0} {
			frame $w.vic
			frame $w.vat
			frame $w.mb
			pack $w.vic $w.vat $w.mb -side right
			set ui [new VicUI $w.vic $local_chan_ $glob_chan_ $videoAgent_ $vpipe_ "$self exit" $vSpec]
			new VatUI $w.vat $glob_chan_ $audioAgent_ "$self exit" $vSpec
		} else {
			import PaneManager
			frame $w.v
			frame $w.v.vic
			frame $w.v.vat
			new PaneManager $w.v.vic $w.v.vat -orient vert -percent 0.75
			frame $w.mb
			new PaneManager $w.v $w.mb -orient horizontal -percent 0.4
			set ui [new VicUI $w.v.vic $local_chan_ $glob_chan_ $videoAgent_ $vpipe_ "$self exit" $vSpec]
			new VatUI $w.v.vat $glob_chan_ $audioAgent_ "$self exit" $vSpec
		}

		set mbui_ [new MBUI $w.mb $mgr_ $sender_ [$self get_option showUI] "$self exit" ]
	}

	pack $w -expand 1 -fill both

	if { [$self get_option useScuba] != "" } {
		$ui set scuba_sess_ $scuba_sess_
	}

	return $ui
}

#
# Parses the sdp message into mui application options, which are then parsed
# as if they had been provided when the MuiApplication was initially instantiated.
# Returns any remaining unprocessed arguments.
# If there is no -sdpFile or -sdp from which to gather an sdp announcement,
# nothing is returned.  The app will then rely on -v -mb -a etc. options.
# Warning: checks are not performed to ensure that -sdp, -sdpFile, or
# {-v, -a, -mb} args are not simultaneously provided to MuiApplication.
# Currently they will just clobber each other.
#
MuiApplication public parse_sdp_msg { } {
	# get the sdp announcement
	set sdp_anncmt_file [$self get_option sdpAnnouncementFile]
	if { $sdp_anncmt_file != "" } {
		# open file and read out sdp announcement
		if [catch {open $sdp_anncmt_file r} fileId] {
			puts stderr "Cannot open $sdp_anncmt_file: $fileId"
		} else {
			append sdp_announcement ""
			while {[gets $fileId line] >= 0} {
				append sdp_announcement "$line\n"
			}
			close $fileId
		}
	} else {
		set sdp_announcement [$self get_option sdpAnnouncement]
	}

	set sdp_msg ""
	# get the 1st sdp message out of the sdp announcement
	if { $sdp_announcement != "" } {
		# parse the announcement into a list of sdp messages
		set parser [new SDPParser 1]		
		set sdp_msgs [$parser parse $sdp_announcement]
		if { [llength $sdp_msgs] != 1 } {
			puts stderr "Invalid SDP announcement:\
					\n$sdp_announcement"
			exit 0
		} else {
			# get the 1st sdp message
			set sdp_msg [lindex $sdp_msgs 0]
		}
	}

	# translate the sdp message into the appropriate MuiApplication options
	# and process these options
	# (modelled after the manner in which nsdr converts an
	#  sdp announcement into options to pass to the mui application)
	if { $sdp_msg != "" } {
		set prog [new Program $sdp_msg]
		set user_app [new UserApplication/MUI]
		set options [$user_app generate_options $prog]
		append argv $options
		# here, options like -v -a -mb etc. should be added to options database
		# and extraneous options are returned
		return [[$self options] parse_args $argv]
	} else {
		return
	}
}

#
# A hook to gracefully shut down mui.  Shuts down the
# underlying audio, video, and mediaboard agents and terminates the application.
#
MuiApplication public exit {} {
	$self instvar videoAgent_ audioAgent_ ui_
        $self instvar mbAgent_ mbui_ mgr_
	$self instvar have_video_ have_audio_ have_mb_
        global mb

	delete $ui_

	if { $have_mb_ } {
		# destroy MediaBoard stuff
		# (agent will delete the session)
		delete $mbAgent_
		delete $mbui_
		delete $mb(gsinterp)
		delete $mgr_
	}
	if { $have_audio_ } {
		$audioAgent_ shutdown
	}
	if { $have_video_ } {
		$videoAgent_ shutdown
	}
	
	delete $self
	exit 0
}

#
# The commandline arguments that can be used for this application. 
#
MuiApplication public init_args { o } {	
	# mui-specific options...
	$o register_option -v videoSessionSpec
	$o register_option -mb mbSessionSpec
	# MPlug will want to pass this app an sdp announcement
	# in the form of a stream of data
	$o register_option -sdp sdpAnnouncement
	# A Helper will want to pass this app an sdp announcement
	# in a file
	$o register_option -sdpFile sdpAnnouncementFile

	# from vic...
	$o register_option -a audioSessionSpec
	$o register_option -B maxbw
	$o register_option -C conferenceName
	$o register_option -c dither
	$o register_option -D device
	$o register_option -af audioFormat
	$o register_option -vf videoFormat
	$o register_option -F maxfps
	$o register_option -I confBusChannel
	$o register_option -K sessionKey
	$o register_option -M colorFile
	$o register_option -m mtu
	$o register_option -o outfile
	$o register_option -q jpegQfactor
	$o register_option -t defaultTTL
	$o register_option -T softJPEGthresh
	$o register_option -U stampInterval
	$o register_option -V visual
	$o register_option -N rtpName
	$o register_list_option -map rtpMap
	
	$o register_boolean_option -H useHardwareDecode
	$o register_boolean_option -P privateColormap

	# MeGa resources
	$o register_option -aofmt megaAudioFormat
	$o register_option -vofmt megaVideoFormat
	$o register_option -mbofmt megaMbFormat

	$o register_option -arport megaRecvAudioPort
	$o register_option -vrport megaRecvVideoPort
	$o register_option -mbrport megaRecvMbPort

	$o register_option -ausemega megaAudioSession
	$o register_option -vusemega megaVideoSession
	$o register_option -mbusemega megaMbSession

	$o register_option -amegactrl megaAudioCtrl
	$o register_option -vmegactrl megaVideoCtrl
	$o register_option -mbmegactrl megaMbCtrl

	$o register_option -asspec audioSessionSpec
	$o register_option -vsspec videoSessionSpec
	$o register_option -mbsspec mbSessionSpec

	$o register_option -amaxsbw maxAudioSessionBW 
	$o register_option -vmaxsbw maxVideoSessionBW 
	$o register_option -mbmaxsbw maxMbSessionBW 

	$o register_option -asbw audioSessionBW 
	$o register_option -vsbw videoSessionBW 
	$o register_option -mbsbw mbSessionBW
	
	$o register_option -asloc audioServiceLocation
	$o register_option -vsloc videoServiceLocation
	$o register_option -mbsloc mbServiceLocation

	$o register_boolean_option -scuba useScuba

	# For MeGa use only.
	$o register_boolean_option -localscuba localScubaScope

	# cues option
	$o register_boolean_option -useCues useCues

	# from vat...
	$o register_boolean_option -r compat
	$o register_option -confid confid   
	$o register_option -loopback loopback
	$o register_list_option -map rtpMap

	# from mb...
	$o register_option -u    uid   
	$o register_option -dbg  debug
	$o register_option -ui   showUI
	$o register_option -drop drop
	$o register_option -play playScript
	$o register_option -rt   rtPlay
	$o register_option -rec  record
	$o register_option -tr   trace
	$o register_option -dp   delayParams
	$o register_option -follow followActive	
	$o register_boolean_option -recvonly recvOnly
	$o register_option -geometry geometry
}

#
# Reset the appropriate objects depending on which media the
# AddressBlock, <i>ab</i>, represents.
#
MuiApplication public reset { ab } {
	$self instvar vspec_ aspec_ mbspec_ ui_
	$self instvar have_video_ have_audio_ have_mb_

	set addr [$ab addr 0]
	if { $have_video_} {
		$self instvar videoHandler_
		$videoHandler_ reset $ab
	}
	if { $have_video_ && [[new AddressBlock $vspec_] addr 0] == $addr } {
		$self instvar videoHandler_
		$videoHandler_ reset $ab
	} elseif { $have_audio_ && [[new AddressBlock $aspec_] addr 0] == $addr } {
	} elseif { $have_mb_ && [[new AddressBlock $mbspec_] addr 0] == $addr } {
		$self instvar mbAgent_ mbui_ mgr_ session_
		set nm [$mbAgent_ set network_]
		$nm loopback 1
		
		# do this for the first time only
		$self instvar have_network_
		if {![info exists have_network_] || $have_network_ == 0} {
			$session_ start_timers
			[$self set sender_] attach $mgr_
			set have_network_ 1
		}

		$mbui_ reset $session_
	}
	
	$ui_ reset
}

#
# Add default font configurations to the options database.
#
MuiApplication public init_fonts { o } {
	new FontInitializer $o

	# from VicApplication...
	option add *Font [$o get_option helv12b] startupFile
	
	# from VatApplication...
	option add *Radiobutton.font [$o get_option helv12b] 100

	# override the default fonts like mb does...
        $o add_option smallfont [$o get_option helv10]	
        $o add_option medfont [$o get_option helv12]
        $o add_option helpFont [$o get_option times12]
}

#
# Initialize the configuration data base with some of the default
# options assumed by vic.  These can be overridden.
#
MuiApplication public init_vic_resources { o } {
        global tcl_platform
	
	# used by TopLevelWindow and for VicUI's window-title
	$o add_default iconPrefix Collaborator:

	# used when creating new CoordinationBus; get and set by AddressBlock
	$o add_default defaultTTL 16

	# WAS used by init_confbus
	$o add_default confBusChannel 0 

	# used by ControlMenu; get and possibly set by AddressBlock
	$o add_default videoFormat h.261
	
	# used by VideoPipeline; set by ControlMenu
	$o add_default useJPEGforH261 false
	
	# used by MetworkMananger; only apps that set this are vic and mui.  maybe it should be add_option
	$o add_default useLayersWindow 1

	# used by c++ code
	$o add_default jvColors 32 
	$o add_default softJPEGthresh -1 
	$o add_default softJPEGcthresh 6 

	# used by VicApplication and ControlMenu
	# (scuba) default session bandwidth (kb/s).
	$o add_default maxVideoSessionBW 128000

        # used by the Info window and ControlMenu
        if {$tcl_platform(platform) != "windows"} {
	        option add Vic.background gray85 startupFile
	        option add Vic.foreground black startupFile
	        option add Vic.disabledForeground gray40 startupFile
	        $o add_default background gray85
                $o add_default foreground black
                $o add_default disabledForeground gray40
        }
}

#
# Initialize the configuration data base with some of the default
# options assumed by vat.  These can be overridden.
#
MuiApplication public init_vat_resources { o } {
	# used when creating new CoordinationBus; get and set by AddressBlock
	$o add_default defaultTTL 16

	# used by c++ code
	$o add_default afDevice -1
	$o add_default afBlocks 2
	$o add_default afSoftOuputGain 0
	$o add_default afSoftInputGain 0
	$o add_default audioFileName /dev/audio

	# used by AudioControlMenu; set by AddressBlock
	$o add_default audioFormat PCM2

	# WAS used by init_confbus
	$o add_default confBusChannel 0
}

#
# Initialize the configuration data base with some of the default
# options assumed by mb.  These can be overridden.
#
MuiApplication public init_mb_resources { o } {
        global env
        $self instvar class_

        option add *Radiobutton.relief flat startupFile
	
        # option add Mb.geometry 700x500 startupFile    

	# used by Object
        $o add_default debug 1

        $o add_default playScript none
        $o add_default rtPlay 0
        $o add_default record 0
	$o add_default showUI 1

        $o add_default trace none
	
	# default added by MBNet, used by AddressBlock
	$o add_default defaultTTL 31	
}

#
# Initialize the configuration data base with some of the default
# options specific to mui.  These can be overridden.
#
MuiApplication public init_mui_resources { o } {
        $o add_default transmitOnStartup 0
	option add Mui.geometry 1000x680 startupFile
}

#
# Create a global bus for communication between tools
# on different machines, and a local bus for tools on
# the same machine. 
#
MuiApplication instproc init_confbus {} {
	#set channel [$self get_option confBusChannel]
	#$self instvar cb_
	#if { $channel != "" && $channel != 0 } {
	#	set cb_ [new CoordinationBus -channel $channel]
	#} else {
	#	set cb_ ""
	#}

        $self instvar local_chan_ glob_chan_
        set local_chan_ [new CoordinationBus -channel 2]

        set ttl [$self get_option defaultTTL]
        set glob_chan_ [new CoordinationBus -channel 3 -ttl $ttl]
}

# (from MBApp)...
MuiApplication instproc init_mb_variables {} {
        global mb 

        $self mbui_init_default
        catch {unset mb}
        
	$self instvar sender_
        if {[$self get_option record]==1} {
                set sender_ [new Tcl_Recorder MB_Sender]
        } else {
                set sender_ [new MB_Sender]
        }
	
	$self instvar mgr_
	set mgr_ [new MB_Manager]
	set pageMgr [new MBPageMgr $mgr_]
	$mgr_ attach_page_manager $pageMgr
	$mgr_ attach_sender $sender_
}

# (from MBApp)...
MuiApplication instproc mbui_init_default { } {
	# using 19 as the priority level, since it's one less than
	# widgetDefault, which is the level used by all WidgetClass objects
	
	option add *Button*HighlightThickness 1 19
	option add *Button*BorderWidth 1 19

	option add *Scrollbar*HighlightThickness 1 19
	option add *Scrollbar*BorderWidth 1 19

	# XXX: a hack to shrink the scrollbar size
	set scrollbar [scrollbar .___scrollbar___]
	set scrollbarWidth [$scrollbar cget -width]
	set scrollbarWidth [expr {($scrollbarWidth * 2)/3}]
	destroy $scrollbar
	option add *Scrollbar*Width $scrollbarWidth 19

	option add *Entry*HighlightThickness 1 19
	option add *Entry*BorderWidth 2 19
	option add *Entry*Background White 19
	option add *Menubutton*HighlightThickness 1 19
	option add *Menubutton*BorderWidth 1 19
}

#
# Checks for extraneous arguments.
#
MuiApplication private check_argv { argv } {
	if { [llength $argv] > 0 } {
		set extra [lindex $argv 0]
		$self fatal "extra arguments (starting with $extra)"
	}
}

#
# Initializes have_<media>_ instvars.
# For now, at least one of -v <videoSessionSpec> -a <audioSessionSpec> and -mb <mediaboardSessionSpec> are expected.
# The app will be exited if none are provided.
#
MuiApplication private check_hostspecs { } {
	$self instvar have_video_ have_audio_ have_mb_

	set have_video_ [ expr { [$self get_option videoSessionSpec] != "" || [$self get_option megaVideoSession] != "" } ]
	set have_audio_ [ expr { [$self get_option audioSessionSpec] != "" || [$self get_option megaAudioSession] != "" } ]
	set have_mb_    [ expr { [$self get_option mbSessionSpec]    != "" || [$self get_option megaMbSession]    != "" } ]

	if { !$have_video_ && !$have_audio_ && !$have_mb_ } {
		$self fatal "a video, audio, or mb destination address is required"
	}
}

#
# Hooks to make MuiApplication embeddable in a web browser.
#
Class MuiApplication/MPlug -superclass MuiApplication

#
# This class can be used to startup regular muiApplication in scrollable MPlug window.
#
MuiApplication/MPlug public init { argv } {
	$self instvar mplug_

	$self start_mplug
	
	# Start up mui embedded in the browser.
	set root [$mplug_ root]	
	$self next $root.[pid] $argv
}

#
# Starts MPlug using a scrollable window.
#
MuiApplication/MPlug public start_mplug {} {
	$self instvar mplug_

	if { ![info exists mplug_] } {
		set mplug_ [new MPlug/ScrolledUI]
	}
}

#
MuiApplication/MPlug private do_error { msg } {
	$self instvar mplug_
	label [$mplug_ root].label -text $msg -justify left -anchor w
	pack [$mplug_ root].label  -fill both -expand 1 -padx 10 -pady 10
	vwait forever
}

Class MuiApplication/MPlug/SDPMime -superclass MuiApplication/MPlug

#
# This class can be used to startup muiApplication in scrollable MPlug window
# when the browser is passing an sdp announcement as a stream to MPlug.
# (i.e. "application/x-sd" mimetype)
#
MuiApplication/MPlug/SDPMime public init { argv } {
	$self instvar mplug_

	$self start_mplug

	# on success, the following function returns NPRES_DONE which is
	# defined to 0 in a .h file
	# XXX: how do I pull in constants defined in .h files into tcl?
	if { [$mplug_ wait_for_stream] } {
		puts stderr "Could not retrieve stream"
		exit
	}

	if { [$mplug_ get mimetype] != "application/x-sd" } {
		$self do_error "MUI cannot handle mimetype: [$mplug_ mimetype]"
	}

	# the stream is expected to be an sdp announcement
	append argv " -sdp {[$mplug_ get stream]} "

	# Start up mui embedded in the browser.
#	set root [$mplug_ root]
#	$self next $root.[pid] $argv
	$self next $argv
}

